@@ -16,6 +16,7 @@ import logging
 import os
 import random
 import string
+from types import TracebackType
 from typing import Any
 from typing import ClassVar
 from typing import ContextManager
@@ -27,6 +28,7 @@ from typing import List
 from typing import Mapping
 from typing import Optional
 from typing import Tuple
+from typing import Type
 from typing import Union
 from typing import cast
 
@@ -48,10 +50,15 @@ def random_key(*,
                redis: Redis,
                prefix: str = 'pottery:',
                length: int = 16,
-               tries: int = 3,
+               num_tries: int = 3,
                ) -> str:
-    if tries <= 0:
+    if not isinstance(num_tries, int):
+        raise TypeError('num_tries must be an int >= 0')
+    elif num_tries < 0:
+        raise ValueError('num_tries must be an int >= 0')
+    elif num_tries <= 0:
         raise RandomKeyError(redis)
+
     all_chars = string.digits + string.ascii_letters
     random_char = functools.partial(random.choice, all_chars)
     suffix = ''.join(cast(str, random_char()) for n in range(length))
@@ -61,7 +68,7 @@ def random_key(*,
             redis=redis,
             prefix=prefix,
             length=length,
-            tries=tries-1,
+            num_tries=num_tries-1,
         )
     return key
 
@@ -70,19 +77,18 @@ class _Common:
     _RANDOM_KEY_PREFIX: ClassVar[str] = 'pottery:'
 
     def __init__(self,
-                 *args: Any,
+                 *,
                  redis: Optional[Redis] = None,
                  key: Optional[str] = None,
-                 **kwargs: Any,
                  ) -> None:
-        self.redis = _default_redis if redis is None else redis
-        self.key = key or self._random_key()
+        self.redis = cast(Redis, redis)
+        self.key = cast(str, key)
 
     def __del__(self) -> None:
         if self.key.startswith(self._RANDOM_KEY_PREFIX):
             self.redis.delete(self.key)
-            _logger.info(
-                "Deleted tmp <%s key='%s'> (went out of scope)",
+            _logger.warning(
+                "Deleted tmp <%s key='%s'> (instance is about to be destroyed)",
                 self.__class__.__name__,
                 self.key,
             )
@@ -97,23 +103,29 @@ class _Common:
 
     @property
     def key(self) -> str:
-        return self._key
+        return self._key  # type: ignore
 
     @key.setter
     def key(self, value: str) -> None:
-        self._key = value or self._random_key()
+        self._key = value or self.__random_key()
 
     def _random_key(self) -> str:
         key = random_key(redis=self.redis, prefix=self._RANDOM_KEY_PREFIX)
-        _logger.info(
+        _logger.warning(
             "Self-assigning tmp key <%s key='%s'>",
             self.__class__.__name__,
             key,
         )
         return key
 
+    # Preserve the Open-Closed Principle with name mangling.
+    #   https://youtu.be/miGolgp9xq8?t=2086
+    #   https://stackoverflow.com/a/38534939
+    __random_key = _random_key
 
-JSONTypes = Union[None, bool, int, float, str, List, Dict[str, Any]]
+
+JSONTypes = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]
+RedisValues = Union[bytes, str, float, int]
 
 
 class _Encodable:
@@ -124,8 +136,8 @@ class _Encodable:
 
     @staticmethod
     def _decode(value: bytes) -> JSONTypes:
-        decoded = json.loads(value.decode('utf-8'))
-        return cast(JSONTypes, decoded)
+        decoded: JSONTypes = json.loads(value.decode('utf-8'))
+        return decoded
 
 
 class _Comparable(metaclass=abc.ABCMeta):
@@ -171,7 +183,27 @@ class _Clearable(metaclass=abc.ABCMeta):
         self.redis.delete(self.key)
 
 
-class Pipelined(metaclass=abc.ABCMeta):
+class _ContextPipeline:
+    def __init__(self, redis: Redis) -> None:
+        self.redis = redis
+
+    def __enter__(self) -> Pipeline:
+        self.pipeline = self.redis.pipeline()
+        return self.pipeline
+
+    def __exit__(self,
+                 exc_type: Optional[Type[BaseException]],
+                 exc_value: Optional[BaseException],
+                 traceback: Optional[TracebackType],
+                 ) -> None:
+        if exc_type is None:
+            with contextlib.suppress(RedisError):
+                self.pipeline.multi()
+                self.pipeline.ping()
+            self.pipeline.execute()
+
+
+class _Pipelined(metaclass=abc.ABCMeta):
     @property
     @abc.abstractmethod
     def redis(self) -> Redis:
@@ -183,50 +215,39 @@ class Pipelined(metaclass=abc.ABCMeta):
         'Redis key.'
 
     @contextlib.contextmanager
-    def _pipeline(self) -> Generator[Pipeline, None, None]:
-        pipeline = self.redis.pipeline()
-        yield pipeline
-        with contextlib.suppress(RedisError):
-            pipeline.multi()
-        pipeline.ping()
-        pipeline.execute()
+    def __watch_keys(self,
+                     *keys: str,
+                     ) -> Generator[Pipeline, None, None]:
+        with _ContextPipeline(self.redis) as pipeline:
+            pipeline.watch(*keys)
+            yield pipeline
 
-    @contextlib.contextmanager
-    def _watch_keys(self,
-                    *keys: Iterable[str],
-                    ) -> Generator[Pipeline, None, None]:
-        original_redis = self.redis
-        keys = keys or (self.key,)
-        try:
-            with self._pipeline() as pipeline:
-                self.redis = pipeline  # type: ignore
-                pipeline.watch(*keys)
-                yield pipeline
-        finally:
-            self.redis = original_redis  # type: ignore
-
-    def _context_managers(self,
-                          *others: Any,
-                          ) -> Generator[ContextManager[Pipeline], None, None]:
+    def __context_managers(self,
+                           *others: Any,
+                           ) -> Generator[ContextManager[Pipeline], None, None]:
         redises = collections.defaultdict(list)
         for container in itertools.chain((self,), others):
-            if isinstance(container, Base):
-                redises[container.redis].append(container)
+            if isinstance(container, _Pipelined):
+                connection_kwargs = frozenset(container.redis.connection_pool.connection_kwargs.items())
+                redises[connection_kwargs].append(container)
         for containers in redises.values():
             keys = (container.key for container in containers)
-            yield containers[0]._watch_keys(*keys)
+            pipeline = containers[0].__watch_keys(*keys)
+            yield pipeline
 
     @contextlib.contextmanager
     def _watch(self,
                *others: Any,
-               ) -> Generator[None, None, None]:
+               ) -> Generator[Pipeline, None, None]:
+        pipelines = []
         with contextlib.ExitStack() as stack:
-            for context_manager in self._context_managers(*others):
-                stack.enter_context(context_manager)
-            yield
+            for context_manager in self.__context_managers(*others):
+                pipeline = stack.enter_context(context_manager)
+                pipelines.append(pipeline)
+            yield pipelines[0]
 
 
-class Base(_Common, _Encodable, _Comparable, _Clearable, Pipelined):
+class Base(_Common, _Encodable, _Comparable, _Clearable, _Pipelined):
     ...
 
 
@@ -244,7 +265,7 @@ class Iterable_(metaclass=abc.ABCMeta):
     @abc.abstractmethod  # pragma: no cover
     def _scan(self,
               *, cursor: int = 0,
-              ) -> Tuple[int, Union[Iterable[JSONTypes], Mapping[JSONTypes, JSONTypes]]]:
+              ) -> Tuple[int, Union[List[bytes], Mapping[bytes, bytes]]]:
         ...
 
     def __iter__(self) -> Generator[JSONTypes, None, None]:
@@ -252,7 +273,7 @@ class Iterable_(metaclass=abc.ABCMeta):
         cursor = 0
         while True:
             cursor, iterable = self._scan(cursor=cursor)
-            yield from (self._decode(cast(bytes, value)) for value in iterable)
+            yield from (self._decode(value) for value in iterable)
             if cursor == 0:
                 break
 
@@ -261,7 +282,8 @@ class Primitive(metaclass=abc.ABCMeta):
     _DEFAULT_MASTERS: ClassVar[FrozenSet[Redis]] = frozenset({_default_redis})
 
     def __init__(self,
-                 *, key: str,
+                 *,
+                 key: str,
                  masters: Iterable[Redis] = frozenset(),
                  ) -> None:
         self.key = key
@@ -278,4 +300,4 @@ class Primitive(metaclass=abc.ABCMeta):
 
     @key.setter
     def key(self, value: str) -> None:
-        self._key = '{}:{}'.format(self.KEY_PREFIX, value)
+        self._key = f'{self.KEY_PREFIX}:{value}'
