@@ -7,16 +7,20 @@
 
 
 import collections
+import contextlib
 import functools
+import itertools
 import logging
 from typing import Any
 from typing import Callable
 from typing import ClassVar
 from typing import FrozenSet
+from typing import Hashable
+from typing import Iterable
 from typing import NamedTuple
 from typing import Optional
-from typing import Tuple
 from typing import TypeVar
+from typing import Union
 from typing import cast
 
 from redis import Redis
@@ -26,6 +30,9 @@ from typing_extensions import Final
 from .base import JSONTypes
 from .base import _default_redis
 from .base import random_key
+from .dict import InitArg
+from .dict import InitIter
+from .dict import InitMap
 from .dict import RedisDict
 
 
@@ -41,7 +48,7 @@ class CacheInfo(NamedTuple):
     currsize: int = 0
 
 
-def _arg_hash(*args: Any, **kwargs: Any) -> int:
+def _arg_hash(*args: Hashable, **kwargs: Hashable) -> int:
     return hash((args, frozenset(kwargs.items())))
 
 
@@ -106,7 +113,7 @@ def redis_cache(*,
         nonlocal redis, key
         if key is None:  # pragma: no cover
             key = random_key(redis=cast(Redis, redis))
-            _logger.info(
+            _logger.warning(
                 "Self-assigning key redis_cache(key='%s') for function %s",
                 key,
                 func.__qualname__,
@@ -115,7 +122,7 @@ def redis_cache(*,
         hits, misses = 0, 0
 
         @functools.wraps(func)
-        def wrapper(*args: Any, **kwargs: Any) -> JSONTypes:
+        def wrapper(*args: Hashable, **kwargs: Hashable) -> JSONTypes:
             nonlocal hits, misses
             hash_ = _arg_hash(*args, **kwargs)
             try:
@@ -130,7 +137,7 @@ def redis_cache(*,
             return return_value
 
         @functools.wraps(func)
-        def bypass(*args: Any, **kwargs: Any) -> JSONTypes:
+        def bypass(*args: Hashable, **kwargs: Hashable) -> JSONTypes:
             hash_ = _arg_hash(*args, **kwargs)
             return_value = func(*args, **kwargs)
             cache[hash_] = return_value
@@ -159,6 +166,16 @@ def redis_cache(*,
     return decorator
 
 
+def _set_expiration(func: F) -> F:
+    @functools.wraps(func)
+    def wrapper(self, *args: Any, **kwargs: Any) -> Any:
+        value = func(self, *args, **kwargs)
+        if self._timeout:
+            self._cache.redis.expire(self._cache.key, self._timeout)
+        return value
+    return cast(F, wrapper)
+
+
 class CachedOrderedDict(collections.OrderedDict):
     '''Redis-backed container that extends Python's OrderedDicts.
 
@@ -173,80 +190,135 @@ class CachedOrderedDict(collections.OrderedDict):
         2. It must map document IDs to hydrated documents
         3. It must cache previously hydrated documents
 
-    Properties 1 and 2 are satisfied by Python's OrderedDict.  However,
-    CachedOrderedDict extends Python's OrderedDict to also satisfy property 3.
+    Properties 1 and 2 are satisfied by Python's collections.OrderedDict.
+    However, CachedOrderedDict extends Python's OrderedDict to also satisfy
+    property 3.
     '''
+
     _SENTINEL: ClassVar[object] = object()
     _NUM_TRIES: ClassVar[int] = 3
 
+    @_set_expiration
     def __init__(self,
                  *,
-                 redis: Optional[Redis] = None,
-                 key: Optional[str] = None,
-                 keys: Tuple[JSONTypes, ...] = tuple(),
+                 redis_client: Optional[Redis] = None,
+                 redis_key: Optional[str] = None,
+                 dict_keys: Iterable[JSONTypes] = tuple(),
                  num_tries: int = _NUM_TRIES,
+                 timeout: Optional[int] = _DEFAULT_TIMEOUT,
                  ) -> None:
         self._num_tries = num_tries
-        redis = _default_redis if redis is None else redis
-        init_cache = functools.partial(RedisDict, redis=redis, key=key)
-        self._cache = self._retry(init_cache)
+        self._timeout = timeout
+        init_cache = functools.partial(
+            RedisDict,
+            redis=redis_client,
+            key=redis_key,
+        )
+        self._cache = self.__retry(init_cache)
         self._misses = set()
 
-        # We have to iterate over keys multiple times, so cast it to a tuple.
-        # This allows the caller to pass in a generator for keys, and we can
-        # still iterate over it multiple times.
-        items, keys = [], tuple(keys)
-        if keys:
-            encoded_keys = (self._cache._encode(key_) for key_ in keys)
-            encoded_values = redis.hmget(self._cache.key, *encoded_keys)
-            for key_, encoded_value in zip(keys, encoded_values):
+        # We have to iterate over dict_keys multiple times, so cast it to a
+        # tuple.  This allows the caller to pass in a generator for dict_keys,
+        # and we can still iterate over it multiple times.
+        items, dict_keys = [], tuple(dict_keys)
+        if dict_keys:
+            encoded_keys = (
+                self._cache._encode(dict_key) for dict_key in dict_keys
+            )
+            encoded_values = self._cache.redis.hmget(
+                self._cache.key,
+                *encoded_keys,
+            )
+            for dict_key, encoded_value in zip(dict_keys, encoded_values):
                 if encoded_value is None:
                     value = self._SENTINEL
-                    self._misses.add(key_)
+                    self._misses.add(dict_key)
                 else:
                     value = self._cache._decode(encoded_value)
-                item = (key_, value)
+                item = (dict_key, value)
                 items.append(item)
         return super().__init__(items)
 
     def misses(self) -> FrozenSet[JSONTypes]:
         return frozenset(self._misses)
 
-    def __setitem__(self, key: JSONTypes, value: JSONTypes) -> None:
+    @_set_expiration
+    def __setitem__(self,
+                    dict_key: JSONTypes,
+                    value: Union[JSONTypes, object]
+                    ) -> None:
+        'Set self[dict_key] to value.'
         if value is not self._SENTINEL:
-            self._cache[key] = value
-            self._misses.discard(key)
-        return super().__setitem__(key, value)
+            self._cache[dict_key] = value
+            self._misses.discard(dict_key)
+        return super().__setitem__(dict_key, value)
 
+    @_set_expiration
     def setdefault(self,
-                   key: JSONTypes,
+                   dict_key: JSONTypes,
                    default: JSONTypes = None,
                    ) -> JSONTypes:
+        '''Insert key with a value of default if key is not in the dictionary.
+
+        Return the value for key if key is in the dictionary, else default.
+        '''
         retriable_setdefault = functools.partial(
-            self._retriable_setdefault,
-            key,
+            self.__retriable_setdefault,
+            dict_key,
             default=default,
         )
-        self._retry(retriable_setdefault)
-        self._misses.discard(key)
-        if key not in self or self[key] is self._SENTINEL:
-            self[key] = default
-        return cast(JSONTypes, self[key])
+        self.__retry(retriable_setdefault)
+        self._misses.discard(dict_key)
+        if dict_key not in self or self[dict_key] is self._SENTINEL:
+            value = self[dict_key] = default
+        else:
+            value = self[dict_key]
+        return value
 
-    def _retriable_setdefault(self,
-                              key: JSONTypes,
-                              default: JSONTypes = None,
-                              ) -> None:
-        with self._cache._watch():
-            if key not in self._cache:
-                self._cache.redis.multi()
-                self._cache[key] = default
+    # Preserve the Open-Closed Principle with name mangling.
+    #   https://youtu.be/miGolgp9xq8?t=2086
+    #   https://stackoverflow.com/a/38534939
+    def __retriable_setdefault(self,
+                               dict_key: JSONTypes,
+                               default: JSONTypes = None,
+                               ) -> None:
+        with self._cache._watch() as pipeline:
+            if dict_key not in self._cache:
+                pipeline.multi()
+                # The following line is equivalent to: self._cache[dict_key] = default
+                pipeline.hset(
+                    self._cache.key,
+                    self._cache._encode(dict_key),
+                    self._cache._encode(default),
+                )
 
-    def _retry(self, callable: Callable[[], Any], try_num: int = 0) -> Any:
+    def __retry(self, callable: Callable[[], Any], try_num: int = 0) -> Any:
         try:
             return callable()
         except WatchError:  # pragma: no cover
             if try_num < self._num_tries - 1:
-                return self._retry(callable, try_num=try_num+1)
+                return self.__retry(callable, try_num=try_num+1)
             else:
                 raise
+
+    @_set_expiration
+    def update(self, arg: InitArg = tuple(), **kwargs: JSONTypes) -> None:  # type: ignore
+        '''D.update([E, ]**F) -> None.  Update D from dict/iterable E and F.
+        If E is present and has an .items() method, then does:  for k in E: D[k] = E[k]
+        If E is present and lacks an .items() method, then does:  for k, v in E: D[k] = v
+        In either case, this is followed by: for k in F:  D[k] = F[k]
+
+        The base class, OrderedDict, has an .update() method that works just
+        fine.  The trouble is that it executes multiple calls to .__setitem__()
+        therefore multiple round trips to Redis.  This overridden .update()
+        makes a single bulk call to Redis.
+        '''
+        to_cache = {}
+        with contextlib.suppress(AttributeError):
+            arg = cast(InitMap, arg).items()
+        for dict_key, value in itertools.chain(cast(InitIter, arg), kwargs.items()):
+            if value is not self._SENTINEL:
+                to_cache[dict_key] = value
+                self._misses.discard(dict_key)
+            super().__setitem__(dict_key, value)
+        self._cache.update(to_cache)
