@@ -13,10 +13,10 @@ from typing import Union
 from typing import cast
 
 from redis import Redis
-from redis.client import Pipeline
 
 from .base import Base
 from .base import JSONTypes
+from .base import RedisValues
 
 
 class HyperLogLog(Base):
@@ -27,10 +27,13 @@ class HyperLogLog(Base):
 
     antirez's blog post:
         http://antirez.com/news/75
+
+    Riak blog post:
+        https://riak.com/posts/technical/what-in-the-hell-is-hyperloglog/index.html?p=13169.html
     '''
 
     def __init__(self,
-                 iterable: Iterable[JSONTypes] = frozenset(),
+                 iterable: Iterable[RedisValues] = frozenset(),
                  *,
                  redis: Optional[Redis] = None,
                  key: Optional[str] = None,
@@ -43,27 +46,38 @@ class HyperLogLog(Base):
         super().__init__(redis=redis, key=key)
         self.update(iterable)
 
-    def add(self, value: JSONTypes) -> None:
+    def add(self, value: RedisValues) -> None:
         'Add an element to a HyperLogLog.  O(1)'
         self.update({value})
 
-    def update(self, *objs: Union['HyperLogLog', Iterable[JSONTypes]]) -> None:
-        objs = (self,) + tuple(objs)
+    def update(self,
+               *objs: Union['HyperLogLog', Iterable[RedisValues]],
+               ) -> None:
+        # We have to iterate over objs multiple times, so cast it to a tuple.
+        # This allows the caller to pass in a generator for objs, and we can
+        # still iterate over it multiple times.
+        objs = tuple(objs)
         other_hll_keys: List[str] = []
-        encoded_values: List[JSONTypes] = []
-        with self._watch(objs[1:]):
+        encoded_values: List[str] = []
+        with self._watch(objs) as pipeline:
             for obj in objs:
                 if isinstance(obj, self.__class__):
-                    other_hll_keys.append(obj.key)
+                    if self.redis.connection_pool == obj.redis.connection_pool:
+                        other_hll_keys.append(obj.key)
+                    else:  # pragma: no cover
+                        raise RuntimeError(
+                            f"can't update {self} with {obj} as they live on "
+                            "different Redis instances/databases"
+                        )
                 else:
                     for value in cast(Iterable[JSONTypes], obj):
                         encoded_values.append(self._encode(value))
-            cast(Pipeline, self.redis).multi()
-            self.redis.pfmerge(self.key, *other_hll_keys)
-            self.redis.pfadd(self.key, *encoded_values)
+            pipeline.multi()
+            pipeline.pfmerge(self.key, *other_hll_keys)
+            pipeline.pfadd(self.key, *encoded_values)
 
     def union(self,
-              *objs: Iterable[JSONTypes],
+              *objs: Iterable[RedisValues],
               redis: Optional[Redis] = None,
               key: Optional[str] = None,
               ) -> 'HyperLogLog':
@@ -82,8 +96,4 @@ class HyperLogLog(Base):
 
     def __repr__(self) -> str:
         'Return the string representation of a HyperLogLog.  O(1)'
-        return '<{} key={} len={}>'.format(
-            self.__class__.__name__,
-            self.key,
-            len(self),
-        )
+        return f'<{self.__class__.__name__} key={self.key} len={len(self)}>'
