@@ -16,16 +16,23 @@ Lua scripting:
 
 import concurrent.futures
 import contextlib
+import logging
 from typing import ClassVar
 from typing import Iterable
+from typing import Optional
+from typing import cast
 
 from redis import Redis
+from redis import RedisError
 from redis.client import Script
-from redis.exceptions import ConnectionError
-from redis.exceptions import TimeoutError
+from typing_extensions import Final
 
 from .base import Primitive
 from .exceptions import QuorumNotAchieved
+from .executor import BailOutExecutor
+
+
+_logger: Final[logging.Logger] = logging.getLogger('pottery')
 
 
 class NextId(Primitive):
@@ -67,31 +74,41 @@ class NextId(Primitive):
     KEY: ClassVar[str] = 'current'
     NUM_TRIES: ClassVar[int] = 3
 
+    _set_id_script: ClassVar[Optional[Script]] = None
+
     def __init__(self,
-                 *, key: str = KEY,
-                 num_tries: int = NUM_TRIES,
+                 *,
+                 key: str = KEY,
                  masters: Iterable[Redis] = frozenset(),
+                 num_tries: int = NUM_TRIES,
                  ) -> None:
         super().__init__(key=key, masters=masters)
+        self.__register_set_id_script()
         self.num_tries = num_tries
-        self._set_id_script = self._register_set_id_script()
-        self._init_masters()
+        self.__init_masters()
 
-    def _register_set_id_script(self) -> Script:
-        master = next(iter(self.masters))
-        set_id_script = master.register_script('''
-            local curr = tonumber(redis.call('get', KEYS[1]))
-            local next = tonumber(ARGV[1])
-            if curr < next then
-                redis.call('set', KEYS[1], next)
-                return next
-            else
-                return nil
-            end
-        ''')
-        return set_id_script
+    # Preserve the Open-Closed Principle with name mangling.
+    #   https://youtu.be/miGolgp9xq8?t=2086
+    #   https://stackoverflow.com/a/38534939
+    def __register_set_id_script(self) -> None:
+        if self._set_id_script is None:
+            _logger.info(
+                'Registering %s._set_id_script',
+                self.__class__.__name__,
+            )
+            master = next(iter(self.masters))
+            self.__class__._set_id_script = master.register_script('''
+                local curr = tonumber(redis.call('get', KEYS[1]))
+                local next = tonumber(ARGV[1])
+                if curr < next then
+                    redis.call('set', KEYS[1], next)
+                    return next
+                else
+                    return nil
+                end
+            ''')
 
-    def _init_masters(self) -> None:
+    def __init_masters(self) -> None:
         with concurrent.futures.ThreadPoolExecutor() as executor:
             for master in self.masters:
                 executor.submit(master.setnx, self.key, 0)
@@ -102,50 +119,69 @@ class NextId(Primitive):
     def __next__(self) -> int:
         for _ in range(self.num_tries):
             with contextlib.suppress(QuorumNotAchieved):
-                next_id = self._current_id + 1
-                self._current_id = next_id
+                next_id = self.__current_id + 1
+                self.__current_id = next_id
                 return next_id
         else:
             raise QuorumNotAchieved(self.masters, self.key)
 
     def __repr__(self) -> str:
-        return '<{} key={} value={}>'.format(
-            self.__class__.__name__,
-            self.key,
-            self._current_id,
+        return (
+            f'<{self.__class__.__name__} key={self.key} '
+            f'value={self.__current_id}>'
         )
 
     @property
-    def _current_id(self) -> int:
-        futures, current_id, num_masters_gotten = set(), 0, 0
+    def __current_id(self) -> int:
         with concurrent.futures.ThreadPoolExecutor() as executor:
+            futures = set()
             for master in self.masters:
-                futures.add(executor.submit(master.get, self.key))
-            for future in concurrent.futures.as_completed(futures):
-                with contextlib.suppress(TimeoutError, ConnectionError):
-                    current_id = max(current_id, int(future.result()))
-                    num_masters_gotten += 1
-        if num_masters_gotten < len(self.masters) // 2 + 1:
-            raise QuorumNotAchieved(self.masters, self.key)
-        else:
-            return current_id
+                future = executor.submit(master.get, self.key)
+                futures.add(future)
 
-    @_current_id.setter
-    def _current_id(self, value: int) -> None:
-        futures, num_masters_set = set(), 0
-        with concurrent.futures.ThreadPoolExecutor() as executor:
+            current_ids = []
+            for future in concurrent.futures.as_completed(futures):
+                try:
+                    current_id = int(future.result())
+                except RedisError as error:
+                    _logger.error(error, exc_info=True)
+                else:
+                    current_ids.append(current_id)
+
+        num_masters_gotten = len(current_ids)
+        quorum = num_masters_gotten >= len(self.masters) // 2 + 1
+        if quorum:
+            return max(current_ids)
+        else:
+            raise QuorumNotAchieved(self.masters, self.key)
+
+    @__current_id.setter
+    def __current_id(self, value: int) -> None:
+        quorum = False
+
+        with BailOutExecutor() as executor:
+            futures = set()
             for master in self.masters:
                 future = executor.submit(
-                    self._set_id_script,
+                    cast(Script, self._set_id_script),
                     keys=(self.key,),
                     args=(value,),
                     client=master,
                 )
                 futures.add(future)
+
+            num_masters_set = 0
             for future in concurrent.futures.as_completed(futures):
-                with contextlib.suppress(TimeoutError, ConnectionError):
+                try:
                     num_masters_set += future.result() == value
-        if num_masters_set < len(self.masters) // 2 + 1:
+                except RedisError as error:
+                    _logger.error(error, exc_info=True)
+                else:
+                    quorum = num_masters_set >= len(self.masters) // 2 + 1
+                    if quorum:  # pragma: no cover
+                        break
+
+        if not quorum:
             raise QuorumNotAchieved(self.masters, self.key)
 
 
