@@ -24,6 +24,7 @@ Lua scripting:
 
 import concurrent.futures
 import contextlib
+import logging
 import os
 import random
 import time
@@ -32,18 +33,24 @@ from typing import ClassVar
 from typing import Iterable
 from typing import Optional
 from typing import Type
+from typing import cast
 
 from redis import Redis
+from redis import RedisError
 from redis.client import Script
-from redis.exceptions import ConnectionError
-from redis.exceptions import TimeoutError
+from typing_extensions import Final
 
 from .base import Primitive
+from .exceptions import ExtendUnlockedLock
 from .exceptions import ReleaseUnlockedLock
 from .exceptions import TooManyExtensions
+from .executor import BailOutExecutor
 from .timer import ContextTimer
 
 
+_logger: Final[logging.Logger] = logging.getLogger('pottery')
+
+
 class Redlock(Primitive):
     '''Distributed Redis-powered lock.
 
@@ -126,6 +133,10 @@ class Redlock(Primitive):
     NUM_EXTENSIONS: ClassVar[int] = 3
     NUM_RANDOM_BYTES: ClassVar[int] = 20
 
+    _acquired_script: ClassVar[Optional[Script]] = None
+    _extend_script: ClassVar[Optional[Script]] = None
+    _release_script: ClassVar[Optional[Script]] = None
+
     def __init__(self,
                  *,
                  key: str,
@@ -135,51 +146,67 @@ class Redlock(Primitive):
                  num_random_bytes: int = NUM_RANDOM_BYTES,
                  ) -> None:
         super().__init__(key=key, masters=masters)
+        self.__register_acquired_script()
+        self.__register_extend_script()
+        self.__register_release_script()
+
         self.auto_release_time = auto_release_time
         self.num_extensions = num_extensions
         self.num_random_bytes = num_random_bytes
 
         self._value = b''
         self._extension_num = 0
-        self._acquired_script = self._register_acquired_script()
-        self._extend_script = self._register_extend_script()
-        self._release_script = self._register_release_script()
 
-    def _register_acquired_script(self) -> Script:
-        master = next(iter(self.masters))
-        acquired_script = master.register_script('''
-            if redis.call('get', KEYS[1]) == ARGV[1] then
-                local pttl = redis.call('pttl', KEYS[1])
-                return (pttl > 0) and pttl or 0
-            else
-                return 0
-            end
-        ''')
-        return acquired_script
+    # Preserve the Open-Closed Principle with name mangling.
+    #   https://youtu.be/miGolgp9xq8?t=2086
+    #   https://stackoverflow.com/a/38534939
+    def __register_acquired_script(self) -> None:
+        if self._acquired_script is None:
+            _logger.info(
+                'Registering %s._acquired_script',
+                self.__class__.__name__,
+            )
+            master = next(iter(self.masters))
+            self.__class__._acquired_script = master.register_script('''
+                if redis.call('get', KEYS[1]) == ARGV[1] then
+                    local pttl = redis.call('pttl', KEYS[1])
+                    return (pttl > 0) and pttl or 0
+                else
+                    return 0
+                end
+            ''')
 
-    def _register_extend_script(self) -> Script:
-        master = next(iter(self.masters))
-        extend_script = master.register_script('''
-            if redis.call('get', KEYS[1]) == ARGV[1] then
-                return redis.call('pexpire', KEYS[1], ARGV[2])
-            else
-                return 0
-            end
-        ''')
-        return extend_script
+    def __register_extend_script(self) -> None:
+        if self._extend_script is None:
+            _logger.info(
+                'Registering %s._extend_script',
+                self.__class__.__name__,
+            )
+            master = next(iter(self.masters))
+            self.__class__._extend_script = master.register_script('''
+                if redis.call('get', KEYS[1]) == ARGV[1] then
+                    return redis.call('pexpire', KEYS[1], ARGV[2])
+                else
+                    return 0
+                end
+            ''')
 
-    def _register_release_script(self) -> Script:
-        master = next(iter(self.masters))
-        release_script = master.register_script('''
-            if redis.call('get', KEYS[1]) == ARGV[1] then
-                return redis.call('del', KEYS[1])
-            else
-                return 0
-            end
-        ''')
-        return release_script
+    def __register_release_script(self) -> None:
+        if self._release_script is None:
+            _logger.info(
+                'Registering %s._release_script',
+                self.__class__.__name__,
+            )
+            master = next(iter(self.masters))
+            self.__class__._release_script = master.register_script('''
+                if redis.call('get', KEYS[1]) == ARGV[1] then
+                    return redis.call('del', KEYS[1])
+                else
+                    return 0
+                end
+            ''')
 
-    def _acquire_master(self, master: Redis) -> bool:
+    def __acquire_master(self, master: Redis) -> bool:
         acquired = master.set(
             self.key,
             self._value,
@@ -188,9 +215,9 @@ class Redlock(Primitive):
         )
         return bool(acquired)
 
-    def _acquired_master(self, master: Redis) -> int:
+    def __acquired_master(self, master: Redis) -> int:
         if self._value:
-            ttl = self._acquired_script(
+            ttl: int = cast(Script, self._acquired_script)(
                 keys=(self.key,),
                 args=(self._value,),
                 client=master,
@@ -199,44 +226,54 @@ class Redlock(Primitive):
             ttl = 0
         return ttl
 
-    def _extend_master(self, master: Redis) -> bool:
-        extended = self._extend_script(
+    def __extend_master(self, master: Redis) -> bool:
+        extended = cast(Script, self._extend_script)(
             keys=(self.key,),
             args=(self._value, self.auto_release_time),
             client=master,
         )
         return bool(extended)
 
-    def _release_master(self, master: Redis) -> bool:
-        released = self._release_script(
+    def __release_master(self, master: Redis) -> bool:
+        released = cast(Script, self._release_script)(
             keys=(self.key,),
             args=(self._value,),
             client=master,
         )
         return bool(released)
 
-    def _drift(self) -> float:
+    def __drift(self) -> float:
         return self.auto_release_time * self.CLOCK_DRIFT_FACTOR + 2
 
-    def _acquire_masters(self) -> bool:
+    def __acquire_masters(self) -> bool:
         self._value = os.urandom(self.num_random_bytes)
         self._extension_num = 0
-        futures, num_masters_acquired = set(), 0
-        with ContextTimer() as timer, \
-             concurrent.futures.ThreadPoolExecutor() as executor:
+        quorum, validity_time = False, 0.0
+
+        with ContextTimer() as timer, BailOutExecutor() as executor:
+            futures = set()
             for master in self.masters:
-                futures.add(executor.submit(self._acquire_master, master))
+                future = executor.submit(self.__acquire_master, master)
+                futures.add(future)
+
+            num_masters_acquired = 0
             for future in concurrent.futures.as_completed(futures):
-                with contextlib.suppress(TimeoutError, ConnectionError):
+                try:
                     num_masters_acquired += future.result()
-            quorum = num_masters_acquired >= len(self.masters) // 2 + 1
-            elapsed = timer.elapsed() - self._drift()
-            validity_time = self.auto_release_time - elapsed
+                except RedisError as error:  # pragma: no cover
+                    _logger.error(error, exc_info=True)
+                else:
+                    quorum = num_masters_acquired >= len(self.masters) // 2 + 1
+                    if quorum:
+                        elapsed = timer.elapsed() - self.__drift()
+                        validity_time = self.auto_release_time - elapsed
+                        break
+
         if quorum and max(validity_time, 0):
             return True
         else:
             with contextlib.suppress(ReleaseUnlockedLock):
-                self.release()
+                self.__release()
             return False
 
     def acquire(self, *, blocking: bool = True, timeout: int = -1) -> bool:
@@ -286,16 +323,18 @@ class Redlock(Primitive):
         if blocking:
             with ContextTimer() as timer:
                 while timeout == -1 or timer.elapsed() / 1000 < timeout:
-                    if self._acquire_masters():
+                    if self.__acquire_masters():
                         return True
                     else:
                         time.sleep(random.uniform(0, self.RETRY_DELAY/1000))
             return False
         elif timeout == -1:
-            return self._acquire_masters()
+            return self.__acquire_masters()
         else:
             raise ValueError("can't specify a timeout for a non-blocking call")
 
+    __acquire = acquire
+
     def locked(self) -> int:
         '''How much longer we'll hold the lock (unless we extend or release it).
 
@@ -322,26 +361,33 @@ class Redlock(Primitive):
             >>> printer_lock_1.release()
 
         '''
-        futures, num_masters_acquired, ttls = set(), 0, []
         with ContextTimer() as timer, \
              concurrent.futures.ThreadPoolExecutor() as executor:
+            futures = set()
             for master in self.masters:
-                futures.add(executor.submit(self._acquired_master, master))
+                future = executor.submit(self.__acquired_master, master)
+                futures.add(future)
+
+            ttls = []
             for future in concurrent.futures.as_completed(futures):
-                with contextlib.suppress(TimeoutError, ConnectionError):
-                    ttl = future.result()
-                    num_masters_acquired += ttl > 0
-                    ttls.append(ttl)
+                try:
+                    ttls.append(future.result())
+                except RedisError as error:  # pragma: no cover
+                    _logger.error(error, exc_info=True)
+
+            num_masters_acquired = sum(1 for ttl in ttls if ttl > 0)
             quorum = num_masters_acquired >= len(self.masters) // 2 + 1
             if quorum:
                 ttls = sorted(ttls, reverse=True)
                 validity_time = ttls[len(self.masters) // 2]
-                validity_time -= round(timer.elapsed() + self._drift())
+                validity_time -= round(timer.elapsed() + self.__drift())
                 return max(validity_time, 0)
             else:
                 return 0
 
-    def extend(self) -> bool:
+    __locked = locked
+
+    def extend(self) -> None:
         '''Extend our hold on the lock (if we currently hold it).
 
         Usage:
@@ -355,7 +401,6 @@ class Redlock(Primitive):
             >>> 8 * 1000 < printer_lock.locked() < 9 * 1000
             True
             >>> printer_lock.extend()
-            True
             >>> 9 * 1000 < printer_lock.locked() < 10 * 1000
             True
             >>> printer_lock.release()
@@ -363,16 +408,28 @@ class Redlock(Primitive):
         if self._extension_num >= self.num_extensions:
             raise TooManyExtensions(self.masters, self.key)
         else:
-            futures, num_masters_extended = set(), 0
-            with concurrent.futures.ThreadPoolExecutor() as executor:
+            quorum = False
+
+            with BailOutExecutor() as executor:
+                futures = set()
                 for master in self.masters:
-                    futures.add(executor.submit(self._extend_master, master))
+                    future = executor.submit(self.__extend_master, master)
+                    futures.add(future)
+
+                num_masters_extended = 0
                 for future in concurrent.futures.as_completed(futures):
-                    with contextlib.suppress(TimeoutError, ConnectionError):
+                    try:
                         num_masters_extended += future.result()
-            quorum = num_masters_extended >= len(self.masters) // 2 + 1
+                    except RedisError as error:  # pragma: no cover
+                        _logger.error(error, exc_info=True)
+                    else:
+                        quorum = num_masters_extended >= len(self.masters) // 2 + 1
+                        if quorum:
+                            break
+
             self._extension_num += quorum
-            return quorum
+            if not quorum:
+                raise ExtendUnlockedLock(self.masters, self.key)
 
     def release(self) -> None:
         '''Unlock the lock.
@@ -390,17 +447,30 @@ class Redlock(Primitive):
             >>> bool(printer_lock.locked())
             False
         '''
-        futures, num_masters_released = set(), 0
-        with concurrent.futures.ThreadPoolExecutor() as executor:
+        quorum = False
+
+        with BailOutExecutor() as executor:
+            futures = set()
             for master in self.masters:
-                futures.add(executor.submit(self._release_master, master))
+                future = executor.submit(self.__release_master, master)
+                futures.add(future)
+
+            num_masters_released = 0
             for future in concurrent.futures.as_completed(futures):
-                with contextlib.suppress(TimeoutError, ConnectionError):
+                try:
                     num_masters_released += future.result()
-        quorum = num_masters_released >= len(self.masters) // 2 + 1
+                except RedisError as error:  # pragma: no cover
+                    _logger.error(error, exc_info=True)
+                else:
+                    quorum = num_masters_released >= len(self.masters) // 2 + 1
+                    if quorum:
+                        break
+
         if not quorum:
             raise ReleaseUnlockedLock(self.masters, self.key)
 
+    __release = release
+
     def __enter__(self) -> 'Redlock':
         '''You can use a Redlock as a context manager.
 
@@ -422,7 +492,7 @@ class Redlock(Primitive):
             >>> states
             [True, False]
         '''
-        self.acquire()
+        self.__acquire()
         return self
 
     def __exit__(self,
@@ -450,14 +520,12 @@ class Redlock(Primitive):
             >>> states
             [True, False]
         '''
-        self.release()
+        self.__release()
 
     def __repr__(self) -> str:
-        return '<{} key={} value={!r} timeout={}>'.format(
-            self.__class__.__name__,
-            self.key,
-            self._value,
-            self.locked(),
+        return (
+            f'<{self.__class__.__name__} key={self.key} '
+            f'value={str(self._value)} timeout={self.__locked()}>'
         )
 
 
